#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-13
# @Filename: operations.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json

from typing import TYPE_CHECKING, ClassVar, Coroutine

from rich.prompt import Confirm

from gort.overwatcher.helpers import get_actor_ping, restart_actors
from gort.tools import decap, get_lvmapi_route, overwatcher_is_running

from .base import BaseRecipe


if TYPE_CHECKING:
    from gort.devices.spec import Spectrograph


__all__ = ["StartupRecipe", "ShutdownRecipe", "CleanupRecipe"]


OPEN_DOME_MESSAGE = """Do not open the dome if you have not checked the following:
* Humidity is below 80%
* Dew point is below the temperature by > 5 degrees (?)
* Wind is below 35 mph
* There is no-one inside the enclosure
* No rain/good conditions confirmed with the Du Pont observers

Du Pont control room:
   (US) +1 626-310-0436
   (Chile) +56 51-2203-609
Slack:
   #lvm-dupont-observing
"""

SHUTDOWN_MESSAGE = """The shutdown recipe has completed.

Please confirm that the dome is closed and the telescopes
are parked by turning on the dome lights with

await g.enclosure.lights.telescope_bright.on()

and checking webcam LVM-TEL06. Then turn off the lights with

await g.enclosure.lights.telescope_bright.off()

If the dome is not closed, please run

await g.enclosure.close(force=True)

If that does not work, please contact the Du Pont observers.

Du Pont control room:
   (US) +1 626-310-0436
   (Chile) +56 51-2203-609
Slack:
   #lvm-dupont-observing
"""


class StartupRecipe(BaseRecipe):
    """Starts the telescopes, runs the calibration sequence, and opens the enclosure."""

    name = "startup"

    async def recipe(
        self,
        open_enclosure: bool = True,
        confirm_open: bool = True,
        focus: bool = True,
    ):
        """Runs the startup sequence.

        Parameters
        ----------
        gort
            The `.Gort` instance to use.
        open_enclosure
            Whether to open the enclosure.
        confirm_open
            If :obj:`True`, asks the user to confirm opening the enclosure.
        focus
            Whether to focus after the enclosure has open.

        """

        self.gort.log.warning("Running the startup sequence.")

        await self.gort.telescopes.home(
            home_telescopes=True,
            home_kms=True,
            home_focusers=True,
            home_fibsel=True,
        )

        self.gort.log.info("Turning off all calibration lamps and dome lights.")
        await self.gort.nps.calib.all_off()
        await self.gort.enclosure.lights.dome_all_off()
        await self.gort.enclosure.lights.spectrograph_room.off()

        self.gort.log.info("Reconnecting AG cameras.")
        await self.gort.ags.reconnect()

        self.gort.log.info("Taking AG darks.")
        await self.gort.guiders.take_darks()

        if open_enclosure:
            if confirm_open:
                self.gort.log.warning(OPEN_DOME_MESSAGE)
                if not Confirm.ask(
                    "Open the dome?",
                    default=False,
                    console=self.gort._console,
                ):
                    return

            self.gort.log.info("Opening the dome ...")
            await self.gort.enclosure.open()

        if open_enclosure and focus:
            self.gort.log.info("Focusing telescopes.")
            await self.gort.guiders.focus()

        self.gort.log.info("The startup recipe has completed.")


class ShutdownRecipe(BaseRecipe):
    """Closes the telescope for the night."""

    name = "shutdown"

    async def recipe(
        self,
        park_telescopes: bool = True,
        additional_close: bool = False,
        disable_overwatcher: bool = False,
        show_message: bool = True,
    ):
        """Shutdown the telescope, closes the dome, etc.

        Parameters
        ----------
        park_telescopes
            Park telescopes (and disables axes). Set to :obj:`False` if only
            closing for a brief period of time. If the dome fails to close with
            ``park_telescopes=True``, it will try again without parking the
            telescopes.
        additional_close
            Issues an additional ``close`` command after the dome is closed.
            This is a temporary solution to make sure the dome is closed
            while we investigate the issue with the dome not fully closing
            sometimes.
        disable_overwatcher
            If :obj:`True`, disables the Overwatcher.
        show_message
            If :obj:`True`, shows a message with instructions on how to confirm
            the dome is closed.

        """

        errored: bool = False

        self.gort.log.warning("Running the shutdown sequence.")

        tasks: list[asyncio.Task | Coroutine] = []

        self.gort.log.info("Turning off all lamps.")
        tasks.append(self.gort.nps.calib.all_off())

        self.gort.log.info("Making sure guiders are idle.")
        tasks.append(self.gort.guiders.stop())

        self.gort.log.info("Closing the dome.")
        tasks.append(self.gort.enclosure.close(mode="normal"))

        if disable_overwatcher:
            self.gort.log.info("Disabling the overwatcher.")
            tasks.append(self.gort.send_command("lvm.overwatcher", "disable --now"))

        for task in tasks:
            try:
                await task
            except Exception as err:
                self.gort.log.error(f"Error during shutdown: {decap(err)}")
                errored = True

        if park_telescopes:
            self.gort.log.info("Parking telescopes for the night.")
            await self.gort.telescopes.park()

        if additional_close:
            self.gort.log.info("Closing the dome again.")
            await asyncio.sleep(5)
            await self.gort.enclosure.close(force=True)

        if show_message:
            self.gort.log.warning(SHUTDOWN_MESSAGE)

        if errored:
            raise RuntimeError("There were errors during the shutdown recipe.")


class CleanupRecipe(BaseRecipe):
    """Stops guiders, aborts exposures, and makes sure the system is ready to go."""

    name = "cleanup"

    async def recipe(self, readout: bool = True, turn_lamps_off: bool = True):
        """Runs the cleanup recipe.

        Parameters
        ----------
        readout
            If the spectrographs are idle and with a readout pending,
            reads the spectrographs.
        turn_lamps_off
            If :obj:`True`, turns off the dome lights and calibration lamps.

        """

        self.gort.log.info("Stopping the guiders.")
        await self.gort.guiders.stop()

        if not (await self.gort.specs.are_idle()):
            extra_sleep: float = 0
            cotasks = []

            for spec in self.gort.specs.values():
                status = await spec.status()
                names = status["status_names"]

                if await spec.is_reading():
                    self.gort.log.warning(f"{spec.name} is reading. Waiting.")
                    cotasks.append(self._wait_until_spec_is_idle(spec))
                    extra_sleep = 10
                elif await spec.is_exposing():
                    self.gort.log.warning(f"{spec.name} is exposing. Aborting.")
                    cotasks.append(spec.abort())
                elif "IDLE" in names and "READOUT_PENDING" in names:
                    msg = f"{spec.name} has a pending exposure."
                    if readout is False:
                        self.gort.log.warning(f"{msg} Aborting it.")
                        cotasks.append(spec.abort())
                    else:
                        self.gort.log.warning(f"{msg} Reading it.")
                        cotasks.append(spec.actor.commands.read())
                        cotasks.append(self._wait_until_spec_is_idle(spec))
                        extra_sleep = 10

            try:
                await asyncio.gather(*cotasks)

                # HACK: lvmscp says the controller is idle before it actually
                # writes the image to disk. If we reset too fast (as we are going
                # to do just after this) that will crash the exposures.
                # I'll fix that in lvmscp (promise) but for now we add a sleep here
                # to allows the images to post-process and write before resetting.
                await asyncio.sleep(extra_sleep)
            except Exception as ee:
                self.gort.log.error(f"Error during cleanup: {decap(ee)}")
                self.gort.log.warning("Resetting the spectrographs.")

        await self.gort.specs.reset(full=True)

        if turn_lamps_off:
            self.gort.log.info("Turning off all calibration lamps and dome lights.")
            await self.gort.nps.calib.all_off()
            await self.gort.enclosure.lights.dome_all_off()

        # Turn off lights in the dome.
        await asyncio.gather(
            self.gort.enclosure.lights.telescope_red.off(),
            self.gort.enclosure.lights.telescope_bright.off(),
        )

        self.gort.log.info("Cleanup recipe has completed.")

    async def _wait_until_spec_is_idle(self, spec: Spectrograph):
        """Waits until an spectrograph is idle."""

        while True:
            if await spec.is_idle():
                return

            await asyncio.sleep(3)


class PreObservingRecipe(BaseRecipe):
    """Prepares the system for observing."""

    name = "pre-observing"

    async def recipe(self, check_actors: bool = True, reboot_ags: bool = False):
        """Runs the pre-observing sequence."""

        if check_actors:
            self.gort.log.info("Checking actors.")
            actor_ping = await get_actor_ping(discard_disabled=True)
            failed_actors = [actor for actor, ping in actor_ping.items() if not ping]

            if len(failed_actors) > 0:
                self.gort.log.warning(f"Failed to ping actors: {failed_actors}.")
                self.gort.log.info("Restarting actors.")
                await restart_actors(list(failed_actors), self.gort)
                await asyncio.sleep(5)
                self.gort.log.info("Restart complete.")
            else:
                self.gort.log.info("All actors are pinging.")

        # Run a clean-up first in case there are any issues with the specs.
        await self.gort.cleanup(readout=False)

        tasks = []

        tasks.append(
            self.gort.telescopes.home(
                home_fibsel=True,
                home_focusers=True,
                home_kms=True,
                home_telescopes=True,
            )
        )
        tasks.append(self.gort.telescopes.park(disable=False, kmirror=False))
        tasks.append(self.gort.ags.reconnect())
        tasks.append(self.gort.specs.expose(flavour="bias"))

        for task in tasks:
            await task

        if reboot_ags:
            self.gort.log.info("Rebooting AG cameras.")
            await self.gort.ags.power_cycle()
            self.gort.log.info("AG cameras rebooted.")

        # Take a dark for the AG cameras here. This is not the ideal time to do it
        # because there's still light, but we want to be sure we have a dark in case
        # something weird happens and the startup recipe is not run before observing.
        # The telescopes are at park and guider.take_darks() uses that position.
        await self.gort.guiders.take_darks()

        # Create the night log if it doesn't exist.
        await get_lvmapi_route("/logs/night-logs/create")

        # Dump the current configuration to the log so that we know what we
        # were using at the start of the night.
        self.gort.log.info("Dumping current configuration to the log.")
        self.gort.log.info(json.dumps(dict(self.gort.config), indent=2))


class PostObservingRecipe(BaseRecipe):
    """Runs the post-observing tasks.

    These include:

    - Closing the dome.
    - Parking the telescopes.
    - Turning off all lamps.
    - Stopping the guiders.
    - Sending the night log email.

    """

    name = "post-observing"

    email_route: ClassVar[str] = "/logs/night-logs/0/email"

    async def recipe(self, send_email: bool = True, force_park: bool = False):
        """Runs the post-observing sequence."""

        tasks = []

        closed = await self.gort.enclosure.is_closed()
        if not closed:
            # Close here with overcurrent because at this point the dome should
            # be close, so this could indicate a problem with the original close.
            tasks.append(self.gort.enclosure.close(mode="overcurrent"))

        parked = [await tel.is_parked() for tel in self.gort.telescopes.values()]
        if force_park or not all(parked):
            tasks.append(self.gort.telescopes.park())

        tasks.append(self.gort.nps.calib.all_off())
        tasks.append(self.gort.guiders.stop())

        for task in tasks:
            try:
                await task
            except Exception as ee:
                self.gort.log.error(f"Error running post-observing task: {ee}")

        if send_email:
            self.gort.log.info("Sending night log email.")
            result = await get_lvmapi_route(
                self.email_route,
                params={"only_if_not_sent": True},
                timeout=30,
            )
            if not result:
                self.gort.log.warning("Night log had already been sent.")

        # Disable the overwatcher.
        if await overwatcher_is_running():
            cmd = await self.gort.send_command("lvm.overwatcher", "disable")
            if cmd.status.did_fail:
                self.gort.log.error("Failed to disable the overwatcher.")
            else:
                self.gort.log.info("Overwatcher has been disabled.")


class RebootAGsRecipe(BaseRecipe):
    """Reboots the AG cameras."""

    name = "reboot-ags"

    async def recipe(self):
        """Power-cycles and reboots the AG cameras."""

        self.gort.log.info("Rebooting AG cameras.")

        self.gort.log.debug("Stopping guiders and waiting for all cameras to be idle.")
        await self.gort.guiders.stop()
        await asyncio.sleep(10)

        if not await self.gort.ags.are_idle():
            self.gort.log.error(
                "Some cameras are not idle. Manually stop the guiders and "
                "ensure that the AG cameras are idle, then run this recipe again."
            )
            return

        self.gort.log.warning("Power-cycling all AG cameras.")
        try:
            await self.gort.ags.power_cycle()
        except Exception as ee:
            self.gort.log.error("Error power-cycling AG cameras", exc_info=ee)
            return

        self.gort.log.info("AG cameras have been power-cycled and are connected.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-01-17
# @Filename: exceptions.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import inspect

from typing import TYPE_CHECKING, ClassVar

from gort.enums import ErrorCode


if TYPE_CHECKING:
    from clu import Command

    from gort.remote import ActorReply, RemoteCommand


def decapitalize_first_letter(s, upper_rest=False):
    return "".join([s[:1].lower(), (s[1:].upper() if upper_rest else s[1:])])


class GortError(Exception):
    """A custom core GortError exception."""

    DEFAULT_ERROR_CODE: ClassVar[ErrorCode] = ErrorCode.UNCATEGORISED_ERROR
    EMIT_EVENT: ClassVar[bool] = True

    def __init__(
        self,
        message: str | None = None,
        error_code: int | ErrorCode | None = None,
        payload: dict = {},
    ):
        try:
            self.error_code = ErrorCode(error_code or self.DEFAULT_ERROR_CODE)
        except ValueError:
            self.error_code = ErrorCode.UNKNOWN_ERROR
            error_code = self.error_code.value

        self.payload = payload
        self.raw_message = message

        prefix = f"Error {self.error_code.value} ({self.error_code.name})"
        if message is not None and message != "":
            message = decapitalize_first_letter(message)
            super().__init__(f"{prefix}: {message}")
        else:
            super().__init__(prefix)


class OverwatcherError(GortError):
    """An error in the overwatcher."""

    pass


class TroubleshooterCriticalError(OverwatcherError):
    """A critical error in the troubleshooter that will shut down the system."""

    def __init__(self, *args, close_dome: bool = False, **kwargs):
        self.close_dome = close_dome
        super().__init__(*args, **kwargs)


class TroubleshooterTimeoutError(OverwatcherError):
    """The troubleshooter timed out while running a recipe."""

    pass


class RemoteCommandError(GortError):
    """An error in a remote command to an actor."""

    EMIT_EVENT = False

    def __init__(
        self,
        message: str | None,
        command: Command,
        remote_command: RemoteCommand,
        reply: ActorReply | None = None,
    ):
        self.command = command
        self.remote_command = remote_command
        self.actor = remote_command._remote_actor.name
        self.reply = reply

        super().__init__(message, error_code=ErrorCode.COMMAND_FAILED)


class InvalidRemoteCommand(GortError):
    """A command that does not exist or cannot be parsed."""

    def __init__(
        self,
        message: str | None,
        command: Command,
        remote_command: RemoteCommand,
    ):
        self.command = command
        self.remote_command = remote_command
        self.actor = remote_command._remote_actor.name

        super().__init__(message, error_code=ErrorCode.COMMAND_FAILED)


class GortTimeoutError(GortError):
    """A timeout error, potentially associated with a timed out remote command."""

    def __init__(
        self,
        message: str | None,
        command: Command | None = None,
        remote_command: RemoteCommand | None = None,
    ):
        self.command = command
        self.remote_command = remote_command
        self.actor = remote_command._remote_actor.name if remote_command else None

        if self.remote_command:
            error_code = ErrorCode.COMMAND_TIMEDOUT
        else:
            error_code = ErrorCode.TIMEOUT

        super().__init__(message, error_code=error_code)


class GortTimeout(GortError):
    """Raised if a timeout occurs."""

    pass


class GortNotImplemented(GortError):
    """A custom exception for not yet implemented features."""

    DEFAULT_ERROR_CODE = ErrorCode.NOT_IMPLEMENTED

    def __init__(self, message: str | None = None):
        message = "This feature is not implemented yet." if not message else message

        super(GortNotImplemented, self).__init__(message)


class GortDeviceError(GortError):
    """A device error, which appends the name of the device to the error message."""

    DEFAULT_ERROR_CODE = ErrorCode.DEVICE_ERROR

    def __init__(
        self,
        message: str | None = None,
        error_code: int | ErrorCode | None = None,
        payload: dict = {},
    ) -> None:
        from gort.gort import GortDevice

        if message is not None:
            stack = inspect.stack()
            f_locals = stack[1][0].f_locals

            if "self" in f_locals:
                obj = f_locals["self"]
                name = getattr(obj, "name", None)
                if issubclass(obj.__class__, GortDevice) and name is not None:
                    message = f"({name}) {message}"

        super().__init__(message, error_code=error_code, payload=payload)


class GortEnclosureError(GortDeviceError):
    """Enclosure-related error."""

    DEFAULT_ERROR_CODE = ErrorCode.ENCLOSURE_ERROR


class GortNPSError(GortDeviceError):
    """NPS-related error."""

    DEFAULT_ERROR_CODE = ErrorCode.NPS_ERROR


class GortGuiderError(GortDeviceError):
    """Guider-related error."""

    DEFAULT_ERROR_CODE = ErrorCode.GUIDER_ERROR


class GortSpecError(GortDeviceError):
    """Spectrograph-related error."""

    DEFAULT_ERROR_CODE = ErrorCode.SPECTROGRAPH_ERROR


class GortAGError(GortDeviceError):
    """AG-related error."""

    DEFAULT_ERROR_CODE = ErrorCode.AG_ERROR


class GortTelescopeError(GortDeviceError):
    """Telescope-related error."""

    DEFAULT_ERROR_CODE = ErrorCode.TELESCOPE_ERROR


class TileError(GortError):
    """An error associated with a `.Tile`."""

    DEFAULT_ERROR_CODE = ErrorCode.SCHEDULER_TILE_ERROR


class GortObserverError(GortError):
    """An error associated with the `.Observer`."""

    DEFAULT_ERROR_CODE = ErrorCode.OBSERVER_ERROR


class GortObserverCancelledError(GortObserverError):
    """An error issued when the observer is cancelled."""

    EMIT_EVENT = False  # Only used to trigger a cancellation of the observing loop.

    pass


class GortWarning(Warning):
    """Base warning for Gort."""

    pass


class GortUserWarning(UserWarning, GortWarning):
    """The primary warning class."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-26
# @Filename: health.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import ClassVar

from lvmopstools.retrier import Retrier
from lvmopstools.utils import Trigger, with_timeout

from clu.tools import CommandStatus

from gort.overwatcher.core import OverwatcherModule, OverwatcherModuleTask
from gort.overwatcher.helpers import get_actor_ping, restart_actors
from gort.overwatcher.overwatcher import Overwatcher
from gort.tools import decap


__all__ = ["HealthOverwatcher"]


class EmitHeartbeatTask(OverwatcherModuleTask["HealthOverwatcher"]):
    """Emits the heartbeat to indicate that the Overwatcher is alive."""

    name = "emit_heartbeat_task"
    keep_alive = True
    restart_on_error = True

    INTERVAL: ClassVar[float] = 5

    def __init__(self):
        super().__init__()

    async def task(self):
        """Sets the overwatcher heartbeat in ``lvmbeat``."""

        while True:
            try:
                await self.emit_heartbeat()
            except Exception as err:
                self.overwatcher.log.error(
                    f"Failed to set overwatcher heartbeat: {decap(err)}",
                    exc_info=err,
                )

            await asyncio.sleep(self.INTERVAL)

    @Retrier(max_attempts=3, delay=1)
    async def emit_heartbeat(self):
        """Emits the heartbeats, with retries."""

        cmd = await self.gort.send_command(
            "lvmbeat",
            "set overwatcher",
            time_limit=5,
        )

        if cmd.status.did_fail:
            if cmd.status & CommandStatus.TIMEDOUT:
                raise RuntimeError("Timed out setting overwatcher heartbeat.")
            raise RuntimeError("Failed to set overwatcher heartbeat.")


class ActorHealthMonitorTask(OverwatcherModuleTask["HealthOverwatcher"]):
    """Monitors the health of actors."""

    name = "actor_health_monitor_task"
    keep_alive = True
    restart_on_error = True

    INTERVAL: ClassVar[float] = 30

    def __init__(self):
        super().__init__()

        self.ping_triggers: dict[str, Trigger] = {}

    async def task(self):
        """Monitors the health of actors."""

        while True:
            actor_status: dict[str, bool] = {}

            try:
                actor_status = await get_actor_ping(
                    discard_disabled=True,
                    discard_overwatcher=True,
                )

                # Update failed_actors. We want allow a grace window of 1 failed ping.
                for actor in actor_status:
                    if actor not in self.ping_triggers:
                        self.ping_triggers[actor] = Trigger(n=2)

                    if actor_status[actor]:
                        self.ping_triggers[actor].reset()
                    else:
                        self.ping_triggers[actor].set()

                # Now check which actors have failed twice (trigger is set).
                failed = [
                    actor
                    for actor, trigger in self.ping_triggers.items()
                    if trigger.is_set()
                ]

                if len(failed) > 0:
                    if self.overwatcher.state.enabled:
                        await self.restart_actors(failed)
                    else:
                        self.log.info(
                            f"Found unresponsible actors: {', '.join(failed)}. "
                            "Not restarting because the Overwatcher is disabled."
                        )

            except Exception as err:
                self.log.error(
                    f"Failed to check actor health: {err}",
                    exc_info=err,
                )

                if len(failed) > 0:
                    await self.overwatcher.shutdown(
                        "actors found unresponsible and cannot be restarted.",
                        disable_overwatcher=True,
                    )

            finally:
                self.module.troubleshooting = False

            await asyncio.sleep(self.INTERVAL)

    @Retrier(max_attempts=2, delay=5)
    async def restart_actors(self, failed_actors):
        """Restarts actors that have failed."""

        ow = self.overwatcher
        self.module.troubleshooting = True

        is_observing = ow.observer.is_observing
        is_calibrating = ow.calibrations.is_calibrating()

        actors_join = ", ".join(failed_actors)
        await self.notify(f"Found unresponsible actors: {', '.join(failed_actors)}.")

        if is_observing:
            if (
                self.gort.specs.last_exposure
                and not self.gort.specs.last_exposure.done()
                and await self.gort.specs.are_reading()
            ):
                await self.notify("Waiting to read exposures before cancelling.")
                await with_timeout(
                    self.gort.specs.last_exposure,  # type: ignore
                    timeout=60,
                    raise_on_timeout=False,
                )

            await ow.observer.stop_observing(
                immediate=True,
                reason=f"Found unresponsible actors: {actors_join}. "
                "Cancelling observations and restarting them.",
            )

        if is_calibrating:
            await self.notify(
                f"Found unresponsible actors: {actors_join}. "
                "Cancelling calibrations and restarting them.",
            )
            await ow.calibrations.cancel()

        await self.notify("Restarting actors.")
        await restart_actors(failed_actors, self.gort)

        for actor in failed_actors:
            self.ping_triggers[actor].reset()

        await self.notify("Actor restart complete. Resuming normal operations.")

        self.module.troubleshooting = False


class HealthOverwatcher(OverwatcherModule):
    """Monitors health."""

    name = "health"
    tasks = [EmitHeartbeatTask(), ActorHealthMonitorTask()]
    delay = 0

    def __init__(self, overwatcher: Overwatcher):
        super().__init__(overwatcher)

        self.troubleshooting: bool = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-05
# @Filename: troubleshooter.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import hashlib
from dataclasses import dataclass
from time import time

from typing import TYPE_CHECKING, TypedDict

from gort.enums import ErrorCode
from gort.exceptions import (
    GortError,
    TroubleshooterCriticalError,
    TroubleshooterTimeoutError,
)
from gort.tools import LogNamespace, decap

from .recipes import TroubleshooterRecipe


if TYPE_CHECKING:
    from gort.overwatcher.overwatcher import Overwatcher


@dataclass
class TroubleModel:
    """Base model for describing a problem to troubleshoot."""

    error: GortError
    error_code: ErrorCode
    hash: int
    message: str | None = None
    handled: bool = False
    tracking_data: ErrorTrackingDict | None = None


class RecipeBook(dict[str, TroubleshooterRecipe]):
    """A dictionary of recipes."""

    def __init__(self, ts: Troubleshooter):
        recipes_sorted = sorted(
            [Recipe(ts) for Recipe in TroubleshooterRecipe.__subclasses__()],
            key=lambda x: x.priority if x.priority is not None else 1000,
        )

        for recipe in recipes_sorted:
            self[recipe.name] = recipe


class ErrorTrackingDict(TypedDict):
    hash: int
    count: int
    last_seen: float
    reset_on_success: bool


class Troubleshooter:
    """Handles troubleshooting for the Overwatcher."""

    def __init__(self, overwatcher: Overwatcher):
        self.overwatcher = overwatcher
        self.notify = overwatcher.notify

        self.recipes = RecipeBook(self)

        self.log = LogNamespace(
            self.overwatcher.gort.log,
            header=f"({self.__class__.__name__}) ",
        )

        self._event = asyncio.Event()
        self._event.set()

        self.error_tracking: dict[int, ErrorTrackingDict] = {}

    def reset(self, clear_all_tracking: bool = False):
        """Resets the troubleshooter to its initial state."""

        self._event.set()

        for hash in self.error_tracking.copy():
            error = self.error_tracking[hash]
            if error["reset_on_success"] or clear_all_tracking:
                error["count"] = 0
                error["last_seen"] = 0

    @property
    def troubleshooting(self) -> bool:
        """Returns ``True`` if the troubleshooter is currently handling an error."""

        return not self._event.is_set()

    async def wait_until_ready(self, timeout: float | None = None):
        """Blocks if the troubleshooter is handling an error until done."""

        try:
            await asyncio.wait_for(self._event.wait(), timeout=timeout)
        except asyncio.TimeoutError:
            raise TroubleshooterTimeoutError("Troubleshooter timed out.")

    def create_hash(self, error: GortError) -> int:
        """Creates a hash for the error to track it."""

        root = f"{error.error_code.name}_{str(error)}"
        return int(hashlib.sha1(root.encode("utf-8")).hexdigest(), 16) % (10**8)

    async def handle(self, error: Exception | str):
        """Handles an error and tries to troubleshoot it.

        Parameters
        ----------
        error
            The error to troubleshoot. This can be an exception (usually an instance
            of :obj:`.GortError`) or a string with the error message.

        """

        if not isinstance(error, (Exception, str)):
            raise RuntimeError("error must be an exception or a string.")

        if isinstance(error, str):
            error = GortError(error, error_code=ErrorCode.UNCATEGORISED_ERROR)
        elif not isinstance(error, GortError):
            error = GortError(str(error), error_code=ErrorCode.UNCATEGORISED_ERROR)

        hash = self.create_hash(error)
        if hash in self.error_tracking:
            self.error_tracking[hash]["count"] += 1
            self.error_tracking[hash]["last_seen"] = time()
        else:
            self.error_tracking[hash] = {
                "hash": hash,
                "count": 1,
                "last_seen": time(),
                "reset_on_success": True,
            }

        error_model = TroubleModel(
            error=error,
            error_code=error.error_code,
            message=str(error),
            hash=hash,
            tracking_data=self.error_tracking[hash],
        )

        await self.notify(
            f"Troubleshooting error of type {error.error_code.name}: "
            f"{error_model.message}",
            level="warning",
        )

        try:
            self._event.clear()

            for recipe in self.recipes.values():
                if recipe == "cleanup":  # This is a last resort recipe
                    continue

                # TODO: for now the first recipe that matches is the only one that runs.
                if recipe.match(error_model):
                    await self.notify(f"Running troubleshooting recipe {recipe.name}.")

                    error_model.handled = await recipe.handle(error_model)
                    if error_model.handled:
                        break

            if error_model.handled:
                return True

            await self.notify(
                "Error could not be handled. Running clean-up recipe.",
                level="warning",
            )
            cleanup = self.recipes["cleanup"]
            await cleanup.handle(error_model)

            return True

        except TroubleshooterCriticalError as err:
            await self.notify(
                "Shutting down due to critical error while "
                f"troubleshooting: {decap(err)}",
                level="critical",
            )
            await self.overwatcher.shutdown(
                disable_overwatcher=True,
                close_dome=err.close_dome,
            )

        finally:
            self.reset()

        return False

    async def run_cleanup(self):
        """Runs the cleanup recipe."""

        cleanup = self.recipes["cleanup"]
        await cleanup.handle()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-11
# @Filename: transforms.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import math
import pathlib
import re

import astropy.coordinates
import astropy.time
import astropy.units
import numpy
import polars
import yaml
from astropy.coordinates import Angle

from gort import config


__all__ = [
    "read_fibermap",
    "offset_to_master_frame_pixel",
    "xy_to_radec_offset",
    "fibre_slew_coordinates",
    "radec_sexagesimal_to_decimal",
    "fibre_to_master_frame",
    "calculate_position_angle",
    "calculate_field_angle",
    "Siderostat",
    "HomTrans",
    "Mirror",
    "wrap_pa_hex",
]


_FIBERMAP_CACHE: polars.DataFrame | None = None


def read_fibermap(
    path: str | pathlib.Path | None = None,
    force_cache: bool = False,
) -> polars.DataFrame:
    """Reads the fibermap file.

    Parameters
    ----------
    path
        Path to the fibermap file. If :obj:`None` uses the path from the
        configuration file.
    force_cache
        If :obj:`True`, forces a re-read of the fibermap file; otherwise reads it
        from the cache if available.

    Returns
    -------
    fibermap
        A Polars DataFrame with the fibermap data.

    """

    global _FIBERMAP_CACHE

    if path is None:
        lvmcore_config = config["services"]["lvmcore"]
        path = pathlib.Path(lvmcore_config["path"]) / lvmcore_config["fibermap"]
    else:
        path = pathlib.Path(path).absolute()

    assert isinstance(path, pathlib.Path)

    if not path.exists():
        raise FileNotFoundError(f"Fibermap file {str(path)} does not exist.")

    if force_cache is False and _FIBERMAP_CACHE is not None:
        return _FIBERMAP_CACHE

    fibermap_y = yaml.load(open(str(path)), Loader=yaml.CFullLoader)

    schema = fibermap_y["schema"]
    cols = [it["name"] for it in schema]
    dtypes = [it["dtype"] if it["dtype"] != "str" else "<U8" for it in schema]

    fibers = polars.from_numpy(
        numpy.array(
            [tuple(fibs) for fibs in fibermap_y["fibers"]],
            dtype=list(zip(cols, dtypes)),
        ),
    )

    # Lower-case some columns.
    fibers = fibers.with_columns(
        polars.col("targettype", "telescope").str.to_lowercase(),
        fibername=polars.col.orig_ifulabel,
    )

    _FIBERMAP_CACHE = fibers

    return _FIBERMAP_CACHE


def offset_to_master_frame_pixel(
    xmm: float | None = None,
    ymm: float | None = None,
    ra: float | None = None,
    dec: float | None = None,
) -> tuple[float, float]:
    """Determines the pixel on master frame coordinates for an offset.

    Parameters
    ----------
    xmm
        The x offset, in mm, with respect to the central fibre in the IFU.
    ymm
        The y offset, in mm, with respect to the central fibre in the IFU.
    ra
        The offset in RA, in arcsec. See :obj:`xy_to_radec_offset` for the
        caveats on this calculations.
    dec
        The offset in declination, in arcsec.


    Returns
    -------
    pixel
        A tuple with the x and z coordinates of the pixel in the master frame.

    Raises
    ------
    ValueError
        If the pixel is out of bounds.

    """

    PIXEL_SIZE = 9  # microns / pixel
    PIXEL_SCALE = 1  # arcsec/pixel

    XZ_0 = (2500, 1000)  # Central pixel in the master frame

    if xmm is not None and ymm is not None:
        if ra is not None or dec is not None:
            raise ValueError("ra/dec cannot be set along with xmm/ymm.")

    elif ra is not None and dec is not None:
        if xmm is not None or ymm is not None:
            raise ValueError("xmm/ymm cannot be set along with ra/dec.")

        xmm = ra * PIXEL_SIZE / PIXEL_SCALE / 1000
        ymm = -dec * PIXEL_SIZE / PIXEL_SCALE / 1000

    else:
        raise ValueError("Not enough inputs supplied.")

    x_mf = xmm * 1000 / PIXEL_SIZE + XZ_0[0]
    y_mf = ymm * 1000 / PIXEL_SIZE + XZ_0[1]

    if x_mf < 0 or x_mf > 2 * XZ_0[0] or y_mf < 0 or y_mf > 2 * XZ_0[1]:
        raise ValueError("Pixel is out of bounds.")

    return (round(x_mf, 1), round(y_mf, 1))


def xy_to_radec_offset(xpmm: float, ypmm: float):
    """Converts offsets in the IFU to approximate RA/Dec offsets.

    .. warning::
        This is an approximate conversion that assumes the IFU is perfectly
        aligned with the AG cameras in the focal plane and that the field
        de-rotation is perfect. It should only be used to determine initial
        offsets for blind telescope slews.

    Parameters
    ----------
    xpmm
        The x offset with respect to the central IFU fibre, in mm, as defined
        in the fibermap.
    ypmm
        As ``xpmm`` for the y offset.

    Returns
    -------
    radec
        RA/Dec offset in arcsec as a tuple. See the warning above for caveats.

    """

    # In the master frame / focal plane, increase x means increasing RA and
    # increasing y means decreasing Dec (i.e., axes are rotated 180 degrees
    # wrt the usual North up, East left).

    PIXEL_SIZE = 9  # microns/pixel
    PIXEL_SCALE = 1  # arcsec/pixel

    x_arcsec = xpmm * 1000 / PIXEL_SIZE * PIXEL_SCALE
    y_arcsec = ypmm * 1000 / PIXEL_SIZE * PIXEL_SCALE

    return (x_arcsec, -y_arcsec)


def fibre_slew_coordinates(
    ra: float,
    dec: float,
    fibre_name: str,
    derotated: bool = True,
) -> tuple[float, float]:
    """Determines the slew coordinates for a fibre.

    This function provides approximate slew coordinates to place
    a target with coordinates ``ra``, ``dec`` on a given fibre. The
    preferred way to use it is to slew the telescope using the
    coordinates returned by this function but then guide on the
    original target coordinates and use the appropriate pixel on the
    master frame to guide on the fibre.

    Parameters
    ----------
    ra,dec
        The RA and Dec of the target to which to slew.
    fibre_name
        The fibre to which to slew the target, with the format
        ``<ifulabel>-<finifu>``.
    derotated
        Whether a k-mirror is derotating the field. If `False`, the
        field rotation for the current time will be used.

    Returns
    -------
    slew_coordinates
        A tuple of RA/Dec coordinates to slew, which should place
        the target near the desired fibre.

    """

    fibermap = read_fibermap()

    if fibre_name not in fibermap["fibername"]:
        raise NameError(f"Fibre {fibre_name} not found in fibermap.")

    fibre = fibermap.row(named=True, by_predicate=polars.col.fibername == fibre_name)

    ra_off, dec_off = xy_to_radec_offset(fibre["xpmm"], fibre["ypmm"])

    if not derotated:
        field_angle = calculate_field_angle(ra, dec, obstime=None)

        # We rotate by -fa_r. I.e., we are "derotating" the sky then
        # calculating the offsets.
        fa_r = -numpy.radians(field_angle)
        rotm = numpy.array(
            [
                [numpy.cos(fa_r), -numpy.sin(fa_r)],
                [numpy.sin(fa_r), numpy.cos(fa_r)],
            ]
        )
        ra_off, dec_off = rotm.dot(numpy.array([ra_off, dec_off]).T)

        # Account for the fact that when we apply the rotation the axes are
        # 180 deg off.
        dec_off = -dec_off

    ra_off = ra_off / 3600.0 / numpy.cos(numpy.radians(dec))
    dec_off = dec_off / 3600.0

    return ra + ra_off, dec + dec_off


def radec_sexagesimal_to_decimal(ra: str, dec: str, ra_is_hours: bool = True):
    """Converts a string of sexagesimal RA and Dec to decimal."""

    ra_match = re.match(r"([+-]?\d+?)[:hd\s](\d+)[:m\s](\d*\.?\d*)", ra)
    if ra_match is None:
        raise ValueError("Invalid format for RA.")

    ra_groups = ra_match.groups()
    ra_deg = float(ra_groups[0]) + float(ra_groups[1]) / 60 + float(ra_groups[2]) / 3600

    if ra_is_hours:
        ra_deg *= 15

    dec_match = re.match(r"([+-]?\d+?)[:hd\s](\d+)[:m\s](\d*\.?\d*)", dec)
    if dec_match is None:
        raise ValueError("Invalid format for Dec.")

    dec_groups = dec_match.groups()
    dec_deg = float(dec_groups[0])

    if dec_deg >= 0:
        dec_deg += float(dec_groups[1]) / 60 + float(dec_groups[2]) / 3600
    else:
        dec_deg -= float(dec_groups[1]) / 60 - float(dec_groups[2]) / 3600

    return ra_deg, dec_deg


def fibre_to_master_frame(fibre_name: str):
    """Returns the xz coordinates in the master frame of a named fibres.

    Parameters
    ----------
    fibre_name
        The fibre for which to calculate the master frame coordinates,
        with the format ``<ifulabel>-<finifu>``.

    Returns
    -------
    pixel
        A tuple with the x and z coordinates of the pixel in the master frame.

    Raises
    ------
    ValueError
        If the pixel is out of bounds.

    """

    XZ_0 = (2500, 1000)  # Central pixel in the master frame

    fibermap = read_fibermap()

    if fibre_name not in fibermap["fibername"]:
        raise NameError(f"Fibre {fibre_name} not found in fibermap.")

    fibre = fibermap.row(named=True, by_predicate=polars.col.fibername == fibre_name)
    xpmm = fibre["xpmm"]
    ypmm = fibre["ypmm"]

    x_mf, z_mf = offset_to_master_frame_pixel(xmm=xpmm, ymm=ypmm)

    if x_mf < 0 or x_mf > 2 * XZ_0[0] or z_mf < 0 or z_mf > 2 * XZ_0[1]:
        raise ValueError("Pixel is out of bounds.")

    return (round(x_mf, 1), round(z_mf, 1))


def calculate_position_angle(ra: float, dec: float, obstime: astropy.time.Time | str):
    """Calculates the position angle seen for a set of coordinates.

    Parameters
    ---------
    ra
        The RA of the centre of the field.
    dec
        The Dec of the centre of the field.
    obstime
        The time of the observation. Either an ISOT string or an astropy time
        object.

    Returns
    -------
    ph
        The position angle. In an image this is the angle between North and the
        y direction, with positive angles being CCW.

    """

    site = astropy.coordinates.EarthLocation.from_geodetic(**config["site"])

    if isinstance(obstime, str):
        obstime_ap = astropy.time.Time(obstime, format="isot")
    else:
        obstime_ap = obstime

    assert isinstance(obstime, astropy.time.Time)

    obstime_ap.location = site

    lst = obstime_ap.sidereal_time("mean")

    ha = lst.deg - ra

    ha_r = numpy.radians(ha)
    lat_r = numpy.radians(site.lat.deg)
    dec_r = numpy.radians(dec)

    par = numpy.arctan2(
        numpy.sin(ha_r),
        numpy.cos(dec_r) * numpy.tan(lat_r) - numpy.sin(dec_r) * numpy.cos(ha_r),
    )

    return numpy.degrees(par)


def calculate_field_angle(
    ra: float,
    dec: float,
    obstime: str | astropy.time.Time | None = None,
):
    """Returns the field angle for a set of coordinates.

    Parameters
    ---------
    ra
        The RA of the centre of the field.
    dec
        The Dec of the centre of the field.
    obstime
        The time of the observation. Either an ISOT string or an astropy time
        object.

    Returns
    -------
    fa
        The field angle. See `.Siderostat.field_angle` for details.

    """

    if isinstance(obstime, str):
        obstime = astropy.time.Time(obstime, format="isot")
    elif obstime is None:
        obstime = astropy.time.Time.now()

    assert isinstance(obstime, astropy.time.Time)

    siderostat = Siderostat()
    target = astropy.coordinates.SkyCoord(ra=ra, dec=dec, unit="deg", frame="icrs")

    return siderostat.field_angle(target, time=obstime)


def wrap_pa_hex(pa: float):
    """Wraps a position angle to the range -30 to 30 degrees.

    Parameters
    ----------
    pa
        The position angle to wrap. Always in degrees in the range 0 to infinity.

    Returns
    -------
    pa_wrap
        The wrapped position angle in the range -30 to 30 degrees.

    """

    # First convert to the range 0 to 360 degrees.
    pa = pa % 360

    # Then wrap to the range 0 to 60 degrees to account for the IFU hexagonal symmetry.
    pa = pa % 60

    # Finally, wrap to the range -30 to 30 degrees.
    if pa > 30:
        pa -= 60

    return pa


class Siderostat:
    """A siderostat of 2 mirrors.

    Adapted from https://github.com/sdss/lvmtipo/blob/main/python/lvmtipo/siderostat.py

    Parameters
    ----------

    zenang
        Zenith angle of the direction of the exit beam (degrees) in the range
        0..180. Default is the design value of the LVMT: horizontal
    azang
        Azimuth angle of the direction of the exit beam (degrees) in the range
        -180..360 degrees, N=0, E=90.
        Ought to be zero for the LCO LVMT where the FP is north of the
        siderostat and 180 for the MPIA test setup where the FP is
        south of the siderostat. The default is the angle for LCO.
    medSign
        Sign of the meridian flip design of the mechanics.
        Must be either +1 or -1. Default is the LCO LVMT design as build (in newer
        but not the older documentation).
    m1m2dist
        Distance between the centers of M1 and M2 in millimeter.
        The default value is taken from ``LVM-0098_Sky Baffle Design``
        of 2022-04-18 by subracting the 84 and 60 cm distance of the
        output pupils to M1 and M2.
    om2_off_ang
        The offset angle which aligns PW motor angle of the M2 axis,
        which is ax0 of the PW GUI, to the angles of the manuscript.
    om1_off_ang
        The offset angle which aligns PW motor angle of the M1 axis,
        which is ax1 of the PW GUI, to the angles of the manuscript, degrees.
        ``om1_off_ang`` and ``om2_off_ang`` are either angles in degrees or
        both ``astropy.coordinates.Angle``.

    """

    def __init__(
        self,
        zenang: float = 90.0,
        azang: float = 0.0,
        medSign: int = -1,
        m1m2dist: float = 240.0,
        om1_off_ang: float | Angle = 118.969,
        om2_off_ang: float | Angle = -169.752,
    ):
        # the vector b[0..2] is the three cartesian coordinates
        # of the beam after leaving M2 in the topocentric horizontal system.
        # b[0] is the coordinate along E, b[1] along N and b[2] up.
        if isinstance(zenang, (int, float)) and isinstance(azang, (int, float)):
            self.b = numpy.zeros((3))
            self.b[0] = math.sin(math.radians(azang)) * math.sin(math.radians(zenang))
            self.b[1] = math.cos(math.radians(azang)) * math.sin(math.radians(zenang))
            self.b[2] = math.cos(math.radians(zenang))
        else:
            raise TypeError("Invalid data types.")

        self.m1m2len = m1m2dist

        if isinstance(medSign, int):
            if medSign in (1, -1):
                self.sign = medSign
            else:
                raise ValueError("Invalid medSign value.")
        else:
            raise TypeError("Invalid medSign data type.")

        # axes orthogonal to beam. box points essentially to the zenith
        # and boy essentially to the East (at LCO) resp West (at MPIA)
        self.box = numpy.zeros((3))
        self.box[0] = 0.0
        self.box[1] = -self.b[2]
        self.box[2] = self.b[1]
        self.boy = numpy.cross(self.b, self.box)

        # The 3x3 B-matrix converts a (x,y,z) vector on the unit spehre
        # which is a direction from the observer to the star in alt-az (horizontal)
        # coordinates (x points East, y to the North and z to the zenith)
        # into a vector on the unit sphere where the two PW motor angles (and offsets)
        # play the role of the azimuth and polar angles.
        # See https://www.mpia.de/~mathar/public/mathar20201208.pdf

        bproj = math.hypot(self.b[1], self.b[2])  # sqrt(by^2+bz^2)

        self.B = numpy.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]], dtype=numpy.double)
        self.B[0][0] = -self.sign * bproj
        self.B[0][1] = self.sign * self.b[0] * self.b[1] / bproj
        self.B[0][2] = self.sign * self.b[0] * self.b[2] / bproj
        self.B[1][0] = 0
        self.B[1][1] = -self.sign * self.b[2] / bproj
        self.B[1][2] = self.sign * self.b[1] / bproj
        self.B[2][0] = self.b[0]
        self.B[2][1] = self.b[1]
        self.B[2][2] = self.b[2]

        if isinstance(om1_off_ang, Angle) and isinstance(om2_off_ang, Angle):
            self.pw_ax_off = [om1_off_ang.radian, om2_off_ang.radian]  # type: ignore
        else:
            self.pw_ax_off = [math.radians(om1_off_ang), math.radians(om2_off_ang)]

    def field_angle(
        self,
        target: astropy.coordinates.SkyCoord,
        time: astropy.time.Time | str | None = None,
    ):
        """Determines the field angle (direction to NCP).

        Parameters
        ----------
        target
            Sidereal target in ra/dec.
        time
            Time of the observation.

        Returns
        -------
        field_angle
            Field angle (direction to NCP) in degrees.

        """

        if isinstance(time, astropy.time.Time):
            pass
        elif isinstance(time, str):
            time = astropy.time.Time(time, format="isot", scale="utc")
        elif time is None:
            time = astropy.time.Time.now()

        # Compute mirror positions
        site = astropy.coordinates.EarthLocation.from_geodetic(**config["site"])

        assert isinstance(astropy.units.Pa, astropy.units.Unit)
        assert isinstance(astropy.units.um, astropy.units.Unit)

        pres = 101300 * math.exp(-site.height.value / 8135.0) * astropy.units.Pa

        altaz = astropy.coordinates.AltAz(
            location=site,
            obstime=time,
            pressure=pres,
            temperature=15 * astropy.units.deg_C,
            relative_humidity=0.5,
            obswl=0.5 * astropy.units.um,
        )

        horiz = target.transform_to(altaz)

        assert isinstance(horiz.az, astropy.coordinates.Angle)
        assert isinstance(horiz.alt, astropy.coordinates.Angle)

        star = numpy.zeros((3))

        # Same procedure as in the construction of b in the Sider ctor,
        # but with 90-zenang=alt
        star[0] = math.sin(horiz.az.radian) * math.cos(horiz.alt.radian)
        star[1] = math.cos(horiz.az.radian) * math.cos(horiz.alt.radian)
        star[2] = math.sin(horiz.alt.radian)

        # unit vector from M2 to M1
        # we're not normalizing to self.m1m2len but keeping the vector
        # m2tom1 at length 1 to simplify the later subtractions to compute
        # normal vectors from other unit vector
        m2tom1 = numpy.cross(star, self.b)
        vlen = numpy.linalg.norm(m2tom1)
        m2tom1 /= self.sign * float(vlen)

        # surface normal to M1 (not normalized to 1)
        m1norm = star - m2tom1
        # the orthogonal distance of the points of M1 to the origin
        # of coordinates are implied by the m1norm direction and
        # the fact that m2tom1 is on the surface. So the homogeneous
        # coordinate equation applied to m2tom1 should yield m2tom1 itself.
        # This requires (m2tom1 . m1norm -d) * n1morm=0 where dots are dot products.
        # the latter dot product actually requires a normalized m1norm
        vlen = numpy.linalg.norm(m1norm)
        m1norm /= vlen
        m1 = Mirror(m1norm, numpy.dot(m2tom1, m1norm))

        # surface normal to M2 (not normalized to 1)
        m2norm = self.b + m2tom1
        m2 = Mirror(m2norm, 0.0)

        # transformation matrix for the 2 reflections individually and in series
        m1trans = m1.to_hom_trans()
        m2trans = m2.to_hom_trans()
        trans = m2trans.multiply(m1trans)

        # for the field angle need a target that is just a little bit
        # more north (but not too little to avoid loss of precision)
        # 10 arcmin = 0.16 deg further to NCP
        targNcp = target.spherical_offsets_by(Angle("0deg"), Angle("0.16deg"))
        horizNcp = targNcp.transform_to(altaz)

        starNcp = numpy.zeros((3))

        # same procedure as in the construction of b in the Sider ctor,
        # but with 90 minus zenith angle=altitude

        assert isinstance(horizNcp.az, astropy.coordinates.Angle)
        assert isinstance(horizNcp.alt, astropy.coordinates.Angle)

        starNcp[0] = math.sin(horizNcp.az.radian) * math.cos(horizNcp.alt.radian)
        starNcp[1] = math.cos(horizNcp.az.radian) * math.cos(horizNcp.alt.radian)
        starNcp[2] = math.sin(horizNcp.alt.radian)

        # image of targNcp while hitting M1
        m1img = trans.apply(m2tom1)

        # image of targNcp before arriving at M1
        star_off_m1 = m2tom1 + starNcp

        starimg = trans.apply(star_off_m1)

        # virtual direction of ray as seen from point after M2
        # no need to normalize this because the atan2 will do...
        # sign was wrong until 2022-11-19: we need to take the direction
        # from the on-axis star (m1img) to the off-axis start (starimg).
        starvirt = starimg - m1img

        # project in a plane orthogonal to  self.b
        cos_fang = numpy.dot(starvirt, self.box)
        sin_fang = numpy.dot(starvirt, self.boy)

        return numpy.degrees(math.atan2(sin_fang, cos_fang))


class HomTrans:
    """A single affine coordinate transformation.

    Represented internally by a 4x4 matrix as a projective.

    From https://github.com/sdss/lvmtipo/blob/main/python/lvmtipo/homtrans.py

    """

    def __init__(self, entries: numpy.ndarray):
        if isinstance(entries, numpy.ndarray):
            self.matr = entries
        else:
            self.matr = numpy.array(entries, numpy.double)

    def multiply(self, rhs: HomTrans | numpy.ndarray):
        """Multiplies by another transformation.

        Parameters
        ----------
        rhs
            The transformation to the right of the multiplication
            sign. So rhs is applied before this transformation.

        Returns
        -------
        hom_trans
         The homogeneous transformation which is the (non-communtative)
         product of self with rhs, representing the consecutive
         application of rhs, then self.

        """

        if isinstance(rhs, HomTrans):
            prod = numpy.matmul(self.matr, rhs.matr)
            return HomTrans(prod)
        elif isinstance(rhs, numpy.ndarray):
            prod = numpy.matmul(self.matr, rhs)
            return HomTrans(prod)

        raise TypeError("Invalid data types")

    def apply(self, rhs: numpy.ndarray):
        """Apply self transformation to a vector of coordinates.

        Parameters
        ----------
        rhs
            The vector. If it has only the standard 3 coordinates,
            a virtual 1 is appended before applying the transformation.

        Returns
        -------
        vector
            A vector of 3 (standard, projected) Cartesian coordinates.

        """

        if isinstance(rhs, numpy.ndarray):
            if rhs.ndim == 1:
                if rhs.shape[0] == 4:
                    prod = numpy.dot(self.matr, rhs)
                elif rhs.shape[0] == 3:
                    w = numpy.append(rhs, [1])
                    prod = numpy.dot(self.matr, w)
                else:
                    raise TypeError("Vector has invalid length.")

                prod /= prod[3]

                return numpy.array([prod[0], prod[1], prod[2]], numpy.double)

            raise TypeError("rhs not  a vector")
        raise TypeError("rhs not numpy array")


class Mirror:
    """A flat mirror.

    This represents an infintely large flat plane. The internal representation is the
    surface normal and the standard equation that the dot product of points on the
    surface by the surface normal equals the distance (of the plane to the origin of
    coordinates).

    From https://github.com/sdss/lvmtipo/blob/main/python/lvmtipo/mirror.py

    Parameters
    ----------
    normal
        The 3 Cartesian coordinates of the surface normal. It must have nonzero
        length, but does not need to be normalized to unit length.
    disttoorg
        The distance of the mirror to the origin of coordinates.
        As in usual geometry, the distance is the shortest distance of the origin
        to the infinitely extended mirror plane.

    """

    def __init__(self, normal: numpy.ndarray, disttoorg: float):
        if isinstance(normal, numpy.ndarray) and isinstance(disttoorg, (int, float)):
            self.d = float(disttoorg)
            if normal.ndim == 1 and normal.shape[0] == 3:
                vlen = numpy.linalg.norm(normal)
                normal /= vlen
                self.n = normal
            else:
                raise TypeError("invalid data types")
        else:
            raise TypeError("invalid data types")

    def to_hom_trans(self):
        """The homogeneous transformation that represents the reflection
        of rays off this mirror surface.

        """

        matr = numpy.zeros((4, 4))

        for r in range(4):
            for c in range(4):
                if r == c:
                    matr[r, c] += 1.0
                if r < 3:
                    if c < 3:
                        matr[r, c] -= 2.0 * self.n[r] * self.n[c]
                    else:
                        matr[r, c] = 2.0 * self.d * self.n[r]

        return HomTrans(matr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-14
# @Filename: telemetry.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from gort.devices.core import GortDevice, GortDeviceSet


if TYPE_CHECKING:
    from gort import ActorReply
    from gort.gort import Gort


__all__ = ["Telemetry", "TelemetrySet"]


class Telemetry(GortDevice):
    """Telemetry sensors."""

    def __init__(self, gort: Gort, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

    async def status(self):
        """Retrieves the status of sensors."""

        reply: ActorReply = await self.actor.commands.status()
        return reply.flatten()


class TelemetrySet(GortDeviceSet[Telemetry]):
    """A set of telemetry sensors."""

    __DEVICE_CLASS__ = Telemetry
    __DEPLOYMENTS__ = ["lvmtelemetry"]
